'' ============================================================
'' 2D Rigid Body Dynamics Simulation Engine
''
'' Copyright (C) 2006 - Jeff Marshall (coder@execulink.com)
''
'' This software is provided 'as-is', without any express or implied
'' warranty. In no event will the authors be held liable for any damages
'' arising from the use of this software.
''	
'' Permission is granted to anyone to use this software for any purpose,
'' including commercial applications, and to alter it and redistribute it
'' freely, subject to the following restrictions:
''	
'' 1. The origin of this software must not be misrepresented; you must not
'' claim that you wrote the original software. If you use this software in
'' a product, an acknowledgment in the product documentation would be
'' appreciated but is not required.
''	
'' 2. Altered source versions must be plainly marked as such, and must not
'' be misrepresented as being the original software.
''	
'' 3. This notice may not be removed or altered from any source
'' distribution.
''
'' ============================================================

Option Explicit
#include "rigid_types.bi"
#include "rigid_bsp.bi"
#include "cons_log.bi"

'' Get the node id of a bsp_node_t
private function getnodeid(node as bsp_node_t ptr) as integer
	function = 0
	if (node) then
		function = node->id
	end if
end function

'' Create a new BSP tree - allocate space for it.
function create_bsp_tree() as bsp_tree_t ptr
	function = callocate(sizeof(bsp_tree_t))
end function

'' Create a new bsp node
function create_bsp_node(byval parent as bsp_node_t ptr, byval rindex as integer, byval perpline1 as integer, byval perpline2 as integer, byval flags as integer, byval dirindex as integer) as bsp_node_t ptr
	dim node as bsp_node_t ptr
	static nodeid as integer = 0
	node = callocate(sizeof(bsp_node_t))
	if (node) then
		nodeid += 1
		node->id = nodeid
		node->parent = parent
		node->linedef_index = rindex
		node->perpline1 = perpline1
		node->perpline2 = perpline2
		node->flags = flags
		node->dirindex = dirindex
	endif
	function = node
end function

'' find a point in a bsp tree's list of point
'' return a zero based index to the point
function find_bsp_point(byval tree as bsp_tree_t ptr, a as vector) as integer
	dim i as integer
	function = -1
	for i = 0 to tree->npoints - 1
		if abs(tree->points[i].x - a.x) < POINT_EPSILON then
			if abs(tree->points[i].y - a.y) < POINT_EPSILON then
				function = i
				exit for
			end if
		end if
	next i
end function

'' Add a new point to a bsp tree's list of points.
'' return an index to the point.
function add_bsp_point(byval tree as bsp_tree_t ptr, a as vector, byval bforce as integer) as integer
	dim i as integer = -1
	if bforce = 0 then
		i = find_bsp_point(tree, a)
	end if
	if i >= 0 then
		function = i
	else
		if tree->npoints + 1 >= tree->maxpoints then
			if tree->maxpoints = 0 then
				tree->maxpoints = 16
			else
				tree->maxpoints *= 2
			end if
			tree->points = reallocate(tree->points, tree->maxpoints * sizeof(vector))
		end if
		function = tree->npoints
		tree->points[tree->npoints] = a
		tree->npoints += 1
	end if
end function

'' find a line def and return the zero based index if found
'' return -1 if not found
function find_bsp_linedef(byval tree as bsp_tree_t ptr, r as line_t) as integer
	dim i as integer
	function = -1
	for i = 0 to tree->nlinedefs - 1
		if tree->linedefs[i].n.x = r.n.x then
			if tree->linedefs[i].n.y = r.n.y then
				if tree->linedefs[i].d   = r.d   then
					function = i
					exit for
				end if
			end if
		end if
	next i
end function

'' add a new linedef to a bsp
function add_bsp_linedef(byval tree as bsp_tree_t ptr, r as line_t, byval bforce as integer) as integer
	dim i as integer = -1
	if bforce = 0 then
		i = find_bsp_linedef(tree, r)
	end if
	if i >= 0 then
		function = i
	else
		if tree->nlinedefs + 1 >= tree->maxlinedefs then
			if tree->maxlinedefs = 0 then
				tree->maxlinedefs = 16
			else
				tree->maxlinedefs *= 2
			end if
			tree->linedefs = reallocate(tree->linedefs, tree->maxlinedefs * sizeof(line_t))
		end if
		function = tree->nlinedefs
		tree->linedefs[tree->nlinedefs] = r
		tree->nlinedefs += 1
	end if
end function

'' check a line segment (ab) against an existing linedef
'' return the two dot products as they are likely can be resused
'' return ENUM:LineDefTypeEnum
function getlinetype(a as vector, b as vector, r as line_t, byref d1 as real, byref d2 as real) as integer
	d1 = VDot(a, r.n) + r.d
	d2 = VDot(b, r.n) + r.d
	if ((d1<0) and (d2>0)) Or ((d1>0) and (d2<0)) then
		function = LT_SPANNING
	elseif ((d1>=0) and (d2>0)) OR ((d1>0) and (d2>=0)) then
		function = LT_IN_FRONT_OF
	elseif ((d1<=0) and (d2<0)) OR ((d1<0) and (d2<=0)) then
		function = LT_IN_BACK_OF
	else
		function = LT_COINCIDENT
	end if
end function

'' partition a bsp tree at the given node. dividing the line segment
'' and call recursively as needed.
function subpartition_bsp(byval tree as bsp_tree_t ptr, byval node as bsp_node_t ptr, a as vector, b as vector, byval rindex as integer, byval perpline1 as integer, byval perpline2 as integer, byval flags as integer, byval dirindex as integer) as integer
	dim lt as integer, m as vector, d1 as real, d2 as real, s as real
	lt = getlinetype(a,b, tree->linedefs[node->linedef_index], d1, d2)
	select case lt
	case LT_IN_BACK_OF
		if node->back then
			function = subpartition_bsp(tree, node->back, a, b, rindex, perpline1, perpline2, flags, dirindex)
		else
			node->back = create_bsp_node(node, rindex, perpline1, perpline2, flags, dirindex)
		end if
	case LT_IN_FRONT_OF
		if node->front then
			function = subpartition_bsp(tree, node->front, a, b, rindex, perpline1, perpline2, flags, dirindex)
		else
			node->front = create_bsp_node(node, rindex, perpline1, perpline2, flags, dirindex)
		end if
	case LT_SPANNING
		'' intersection point = a + ab ( d1 / ( d1 - d2 ))
		s = d1 / (d1 - d2)
		m.x = a.x + (b.x - a.x) * s
		m.y = a.y + (b.y - a.y) * s
		function = subpartition_bsp(tree, node, a, m, rindex, perpline1, perpline2, flags, dirindex)
		function = subpartition_bsp(tree, node, m, b, rindex, perpline1, perpline2, flags, dirindex)
	case LT_COINCIDENT	
		'' FIXME: we shouldn't really do this, we should have leaves at each node with a list
		'' of linedefs instead of duplicating splitting planes like this
		VSub m, b, a
		VPerpIm m
		if VDot(m, tree->linedefs[rindex].n) < 0 then
			if node->back then
				function = subpartition_bsp(tree, node->back, a, b, rindex, perpline1, perpline2, flags, dirindex)
			else
				node->back = create_bsp_node(node, rindex, perpline1, perpline2, flags, dirindex)
			end if
		else
			if node->front then
				function = subpartition_bsp(tree, node->front, a, b, rindex, perpline1, perpline2, flags, dirindex)
			else
				node->front = create_bsp_node(node, rindex, perpline1, perpline2, flags, dirindex)
			end if
		end if
	end select
end function

'' entry point for partitioning a bsp tree with a new line segment
'' this could be combined with subpartition_bsp but is not as the
'' code in this sub only ever needs to be called once at the
'' very start of the subpartion procedure
function partition_bsp(byval tree as bsp_tree_t ptr, a as vector, b as vector, r as line_t, flags as integer, dr as line_t) as integer
	dim as integer rindex, perpline1, perpline2, dirindex
	dim as line_t aface, bface
	dim as vector t
	function = False
	rindex = add_bsp_linedef(tree, r, False)
	dirindex = add_bsp_linedef(tree, dr, False)

	VPerp t, r.n
	VLineEqFromPointNormal bface, b, t
	perpline2 = add_bsp_linedef(tree, bface, False)

	VNegIm t
	VLineEqFromPointNormal aface, a, t
	perpline1 = add_bsp_linedef(tree, aface, False)

	if tree->root = 0 then
		tree->root = create_bsp_node(0, rindex, perpline1, perpline2, flags, dirindex) '' no parent - this is the root node
		function = True
	else
		function = subpartition_bsp(tree, tree->root, a, b, rindex, perpline1, perpline2, flags, dirindex)
	end if
end function

'' delete a bsp tree by walking all the nodes and releasing
'' all memory associated with the tree
sub delete_bsp_tree(byval tree as bsp_tree_t ptr)
	dim as bsp_node_t ptr node, parent
	if (tree) then
		if (tree->root) then
			node = tree->root
			while (node)
				if (node->back) then
					node = node->back
				elseif (node->front) then
					node = node->front
				else
					parent = node->parent
					if (parent) then
						if parent->front = node then parent->front = 0
						if parent->back  = node then parent->back = 0
					end if
					node = parent
				end if
			wend
			deallocate tree->root
		end if
		if tree->linedefs then
			deallocate tree->linedefs
		end if
		if tree->points then
			deallocate tree->points
		end if
		deallocate tree
	end if
end sub

'' dump information about a node in the tree to the console (if enabled)
sub bsp_dump_node(byval tree as bsp_tree_t ptr, byval node as bsp_node_t ptr)
	dim r as line_t
	if (node) then
		Log_Print("Node " & str(node->id))
		r = tree->linedefs[node->linedef_index]
		Log_Print("  Lindef := " & str(node->linedef_index) & " := " & VString(r.n) & "," & str(r.d))
		r = tree->linedefs[node->perpline1]
		Log_Print("  Perp1  := " & str(node->perpline1) & " := " & VString(r.n) & "," & str(r.d))
		r = tree->linedefs[node->perpline2]
		Log_Print("  Perp2  := " & str(node->perpline2) & " := " & VString(r.n) & "," & str(r.d))
		Log_Print("  Parent := " & str(getnodeid(node->parent)))
		Log_Print("  Back   := " & str(getnodeid(node->back)))
		Log_Print("  Front  := " & str(getnodeid(node->front)))
		bsp_dump_node tree, node->back
		bsp_dump_node tree, node->front
	end if
end sub

'' dump information about the tree and all nodes to the console (if enabled)
sub bsp_dump_tree(byval tree as bsp_tree_t ptr)
	dim i as integer
	dim r as line_t
	dim node as bsp_node_t ptr
	Log_Print("Line Defs : " & str(tree->nlinedefs))
	for i = 0 to tree->nlinedefs-1
		r = tree->linedefs[i]
		Log_Print(str(i) & " := " & VString(r.n) & "," & str(r.d))
	next i
	bsp_dump_node tree, tree->root
end sub

'' find the "side" of the world a point lies on.
'' 1 for inside, -1 for outside
function bsp_get_side(byval tree as bsp_tree_t ptr, a as vector) as integer
	dim as bsp_node_t ptr node
	dim d as real
	dim ret as integer
	ret = 0
	node = tree->root
	while (node)	
		d = VDot(a, tree->linedefs[node->linedef_index].n) + tree->linedefs[node->linedef_index].d
		if d < 0 then
			if (node->flags and BSP_SOLID) <> 0 then
				ret = -1
			end if
			node = node->back
		else
			if (node->flags and BSP_SOLID) <> 0 then
				ret = 1
			end if
			node = node->front
		end if
	wend
	function = ret	
end function

function bsp_get_force(byval tree as bsp_tree_t ptr, a as vector, f as vector) as integer
	dim as bsp_node_t ptr node
	dim d as real
	dim ret as integer
	ret = 0
	node = tree->root
	while (node)	
		d = VDot(a, tree->linedefs[node->linedef_index].n) + tree->linedefs[node->linedef_index].d
		if d < 0 then
			if (node->flags and BSP_FORCE) <> 0 then
				f = tree->linedefs[node->dirindex].n
				ret = -1
			end if
			node = node->back
		else
			if (node->flags and BSP_FORCE) <> 0 then
				f = tree->linedefs[node->dirindex].n
				ret = 1
			end if
			node = node->front
		end if
	wend
	function = ret	
end function

'' find a collision with the bsp tree
'' return -1 if the point is outside (too far)
'' return  1 if the point is inside (no collision)
'' return  0 if there is a collision (within EDGE_EPSILON) of a
''         front edge
function bsp_find_collision(byval tree as bsp_tree_t ptr, a as vector, dcol as real, colnrm as vector) as integer
	dim as bsp_node_t ptr node
	dim ret as integer
	dim as vector n, ntmp
	dim as real d, dtmp

	'' TODO: should be able to combine bsp_get_side() and bsp_find_collision
	'' so that only one pass of the tree is needed

	ret = bsp_get_side(tree, a)
	if (ret = 1) then
		node = tree->root
		while (node)	
			n = tree->linedefs[node->linedef_index].n
			d = VDot(a, n) + tree->linedefs[node->linedef_index].d

			if (node->flags and BSP_SOLID) <> 0 then
				if d < 0 then
					node = node->back
				elseif d < EDGE_EPSILON then
					ntmp = tree->linedefs[node->perpline1].n
					dtmp = VDot(a,ntmp) + tree->linedefs[node->perpline1].d - EDGE_EPSILON
					if (dtmp >= 0) then
						ntmp = tree->linedefs[node->perpline2].n
						dtmp = VDot(a,ntmp) + tree->linedefs[node->perpline2].d - EDGE_EPSILON
						if dtmp >= 0 then
							ret = 0
							dcol = d
							colnrm = n
							exit while
						end if
					end if
					if d < 0 then
						node = node->back
					else
						node = node->front
					end if
				else
					node = node->front
				end if
			else
				if d < 0 then
					node = node->back
				else
					node = node->front
				end if
			end if
		wend
	end if
	function = ret	
end function

function bsp_check_visibility(byval tree as bsp_tree_t ptr, byval node as bsp_node_t ptr, a as vector, b as vector) as integer
	dim lt as integer, m as vector, d1 as real, d2 as real, s as real, result as integer
	if (node) then
		lt = getlinetype(a,b, tree->linedefs[node->linedef_index], d1, d2)
		select case lt
		case LT_IN_BACK_OF
			if node->back then
				function = bsp_check_visibility(tree, node->back, a, b)
			else
				function = -1
			end if
		case LT_IN_FRONT_OF, LT_COINCIDENT
			if node->front then
				function = bsp_check_visibility(tree, node->front, a, b)
			else
				function = 1
			end if
		case LT_SPANNING
			'' intersection point = a + ab ( d1 / ( d1 - d2 ))
			s = d1 / (d1 - d2)
			m.x = a.x + (b.x - a.x) * s
			m.y = a.y + (b.y - a.y) * s
			result = bsp_check_visibility(tree, node, a, m)
			if result >= 0 then
				result = bsp_check_visibility(tree, node, m, b)
			end if
			function = result
		end select
	else
		function = -1
	end if
end function
