'' ============================================================
'' 2D Rigid Body Dynamics Simulation Engine
''
'' Copyright (C) 2006 - Jeff Marshall (coder@execulink.com)
''
'' This software is provided 'as-is', without any express or implied
'' warranty. In no event will the authors be held liable for any damages
'' arising from the use of this software.
''	
'' Permission is granted to anyone to use this software for any purpose,
'' including commercial applications, and to alter it and redistribute it
'' freely, subject to the following restrictions:
''	
'' 1. The origin of this software must not be misrepresented; you must not
'' claim that you wrote the original software. If you use this software in
'' a product, an acknowledgment in the product documentation would be
'' appreciated but is not required.
''	
'' 2. Altered source versions must be plainly marked as such, and must not
'' be misrepresented as being the original software.
''	
'' 3. This notice may not be removed or altered from any source
'' distribution.
''
'' ============================================================

Option Explicit

#include "rigid_types.bi"
#include "fbgfx.bi"
#include "vector.bi"
#include "rigid_shapes.bi"
#include "rigid_world.bi"
#include "rigid_engine.bi"
#include "rigid_render.bi"
#include "rigid_bsp.bi"
#include "cons_log.bi"

dim shared AGRAVITY as vector
dim shared as integer STATE1 = 0, STATE2 = 1

dim shared MAX_OBJECTS as integer = 16
dim shared NUM_OBJECTS as integer = 0
redim shared objects(1 to MAX_OBJECTS) as object_t ptr

dim shared MAX_SPRINGS as integer = 16
dim shared NUM_SPRINGS as integer = 0
redim shared springs(1 to MAX_SPRINGS) as spring_t ptr

dim shared MAX_COLLISIONS as integer = 16
dim shared collisions(1 to MAX_COLLISIONS) as collision_t
dim shared collisionstate as collision_state_t

dim shared DEFAULT_SPRING_KS as real
dim shared DEFAULT_SPRING_KD as real
dim shared DEFAULT_COR as real 

'' ==========================================================================
'' DEBUG
'' ==========================================================================

Sub Dbg_Print_Collision(collision as collision_t)
	dim x as string

	with collision
		select case collision.kind
		case CollisionType_WorldToObject: x = "CollisionType_WorldToObject"
		case CollisionType_ObjectToWorld: x = "CollisionType_ObjectToWorld"
		case CollisionType_ObjectToObject: x = "CollisionType_ObjectToObject"
		end select

		Dbg_Print ("  state=" & str(.state) & ", kind=" & str(.kind) & " - " & x)
		Dbg_Print ("  objectindex1=" & str(.objectindex1) & ", pointindex1=" & str(.pointindex1) )
		Dbg_Print ("  objectindex2=" & str(.objectindex2) & ", normal=" & VString(.normal) )
	end with

End Sub


'' ==========================================================================
'' SPRING CREATION
'' ==========================================================================

'' Create and elastic (allocate space for it)
function Create_Spring() as spring_t ptr
	dim i as integer
	for i = 1 to NUM_SPRINGS
		if springs(i)->Alive = False then
			exit for
		end if
	next i
	if i > NUM_SPRINGS then
		NUM_SPRINGS += 1
		if NUM_SPRINGS > MAX_SPRINGS then
			MAX_SPRINGS = MAX_SPRINGS * 2
			redim preserve springs(1 to MAX_SPRINGS) as spring_t ptr
		end if
		springs(i) = callocate(sizeof(spring_t))
	end if
	springs(i)->alive = True
	function = springs(i)
end function

sub Destroy_All_Springs()
	dim as integer i
	dim as spring_t ptr spr
	for i = 1 to NUM_SPRINGS
		spr = springs(i)
		if spr then deallocate spr
	next i
	NUM_SPRINGS = 0
end sub

'' ==========================================================================
'' OBJECT MANAGEMENT
'' ==========================================================================

'' create a new object - allocate space for it
function Create_object() as object_t ptr
	dim obj as object_t ptr, objst as objectstate_t ptr
	obj = callocate(sizeof(object_t))
	if (obj) then
		obj->shp = 0
		obj->mass = 10
		obj->moi = 100
		obj->cor = default_cor '' was 1.0
		VSet obj->appliedforce, 0, 0
		obj->appliedtorque = 0

		objst = @obj->state(0)

		VSet objst->p, 0, 0
		VSet objst->v, 0, 0
		objst->angp = 0
		objst->angv = 0
		VSet objst->force,0,0
		objst->torque = 0
		
		obj->state(1) = obj->state(0)

	end if
	function = obj
end function

'' add an object to the global list of objects
function Add_Object(byval obj as object_t ptr) as integer
	function = False
	if NUM_OBJECTS = MAX_OBJECTS then
		MAX_OBJECTS = MAX_OBJECTS * 2
		redim preserve objects(1 to MAX_OBJECTS) as object_t ptr
	end if
	NUM_OBJECTS = NUM_OBJECTS + 1
	obj->state(0).wpoints = Callocate(obj->shp->npoints * sizeof(vector))
	obj->state(1).wpoints = Callocate(obj->shp->npoints * sizeof(vector))
	objects(NUM_OBJECTS) = obj
	function  = NUM_OBJECTS
end function

sub Destroy_All_Objects()
	dim as integer i
	dim as object_t ptr obj
	for i = 1 to NUM_OBJECTS
		obj = objects(i)
		if obj->state(0).wpoints then deallocate obj->state(0).wpoints
		if obj->state(1).wpoints then deallocate obj->state(1).wpoints
		if obj then	deallocate obj
	next i
	NUM_OBJECTS = 0
end sub

'sub ApplyForceToObject(byval objectindex as integer, p as vector)
'	dim obj as object_t ptr
'	dim objst as objectstate_t ptr
'	dim as vector t
'	obj = objects(objectindex)
'	objst = @obj->state(STATE1)
'	VSub t, p, objst->p
'	VSubIm t, objst->v
'	VMul obj->appliedforce, obj->mass, t
'	VSubIm obj->appliedforce, AGRAVITY
'end sub

sub ApplyForceToObject(byval objectindex as integer, p as vector, byval t as real)
	dim obj as object_t ptr
	dim objst as objectstate_t ptr
	dim as vector t1, t2
	obj = objects(objectindex)
	objst = @obj->state(STATE1)
	VSub t1, p, objst->p
	VMulIm t1, 10
	VSubIm t1, objst->v
	VMul obj->appliedforce, obj->mass, t1
	VSubIm obj->appliedforce, AGRAVITY
	obj->appliedtorque = t * obj->moi
end sub

'' ==========================================================================
'' COLLISION MANAGEMENT
'' ==========================================================================

sub Reset_Collisions()
	dim as integer i
	dim as object_t ptr obj

	Dbg_Print ("Reset_Collisions()")

	with collisionstate
		.count = 0
		.state = CollisionState_None
	end with

	for i = 1 to NUM_OBJECTS
		obj = objects(i)
		obj->colcount = 0
		obj->colworld = 0
		obj->colobject = 0
		obj->immovable = 0
		obj->unresolved = 0
	next i
end sub


'' colstate ENUM:CollisionStateEnum
function Add_Collision(collision as collision_t) as integer
	collisionstate.count += 1
	if collisionstate.count > MAX_COLLISIONS then
		MAX_COLLISIONS *= 2
		redim preserve collisions(1 to MAX_COLLISIONS) as collision_t
	end if
	if collision.state > collisionstate.state then
		collisionstate.state = collision.state
	end if
	collisions(collisionstate.count) = collision
	function = collisionstate.count

	Dbg_Print("Add_Collision(count=" & str(collisionstate.count) & " )")
	Dbg_Print_Collision(collision)
end function

'' ==========================================================================
'' VECTOR MAPPING BETWEEN SHAPES/OBJECTS/WORLD
'' ==========================================================================

'' Map a point on the shape to the world
sub MapShapePointToWorld(v as vector, byval objst as objectstate_t ptr, a as vector)
	v.x = a.x * objst->cs - a.y * objst->sn
	v.y = a.x * objst->sn + a.y * objst->cs
	VAddIm v, objst->p
end sub

'' Map a point in the world to a shape
sub MapWorldPointToShape(v as vector, byval objst as objectstate_t ptr, a as vector)
	dim t as vector
	VSub t, a, objst->p
	v.x = t.x * objst->cs + t.y * objst->sn
	v.y = -t.x * objst->sn + t.y * objst->cs
end sub

'' "Draw" objects in to the world
sub RenderObjectToWorld(byval obj as object_t ptr, byval objst as objectstate_t ptr)
	dim i as integer, j as integer, a as vector
	dim v as vector
	objst->cs = cos(objst->angp)
	objst->sn = sin(objst->angp)
	for i = 0 to obj->shp->npoints - 1
		MapShapePointToWorld objst->wpoints[i], objst, obj->shp->lpoints[i]
	next i
end sub

'' "Draw" all objects in to the world
sub RenderObjectsToWorld(byval stateindex as integer)
	dim i as integer, spr as spring_t ptr
	for i = 1 to NUM_OBJECTS
		RenderObjectToWorld objects(i), @objects(i)->state(stateindex)
	next i
end sub

'' ==========================================================================

sub computeforces(byval stateindex as integer)
	dim as integer i
	dim as vector t, v, springv, springf, relvel, dampf
	dim as vector relpnt1, relperp1, vel1
	dim as vector relpnt2, relperp2, vel2
	dim as object_t ptr obj
	dim as objectstate_t ptr objst
	dim as spring_t ptr spr

	for i = 1 to NUM_OBJECTS
		obj = objects(i)
		objst = @obj->state(stateindex)

		VSet objst->force, 0, 0
		objst->torque = 0

		VMul t, obj->mass, AGRAVITY
		VAddIm objst->force, t         '' Add force due to gravity

		VAddIm objst->force, obj->appliedforce
		objst->torque += obj->appliedtorque

		if GetWorldForceOnPoint(objst->p, t) = -1 then
			VMulIm t, obj->mass '' TODO: make an option to mul force by mass
			VAddIm objst->force, t
		end if

	next i

	for i = 1 to NUM_SPRINGS
		spr = springs(i)		
		if spr->Alive then
			if spr->ObjectIndex1 > 0 then
				'' point attached to object
				objst = @objects(spr->ObjectIndex1)->state(stateindex)
				MapShapePointToWorld spr->wpnt1, objst, spr->pnt1
				VSub relpnt1, spr->wpnt1, objst->p
				VPerp relperp1, relpnt1
				VMul   vel1, objst->angv, relperp1
			else
				'' point is fixed
				spr->wpnt1 = spr->pnt1
				VSet relpnt1, 0, 0
				VSet relperp1, 0, 0
				VSet vel1, 0, 0
			end if

			if spr->ObjectIndex2 > 0 then
				'' point attached to object
				objst = @objects(spr->ObjectIndex2)->state(stateindex)
				MapShapePointToWorld spr->wpnt2, objst, spr->pnt2
				VSub relpnt2, spr->wpnt2, objst->p
				VPerp relperp2, relpnt2
				VMul   vel1, objst->angv, relperp2
			else
				'' point is fixed
				spr->wpnt2 = spr->pnt2
				VSet relpnt2, 0, 0
				VSet relperp2, 0, 0
				VSet vel2, 0, 0
			end if

			'' compute direction of spring and force of spring
			VSub springv, spr->wpnt1, spr->wpnt2
			VMul springf, -spr->springKs, springv

			'' compute relative velocity
			VSub relvel, vel1, vel2

			'' Damping Force - project relvel on to springv
			VMul dampf, -spr->springKd * VDot(relvel, springv) / VDot(springv, springv), springv

			VAddIm springf, dampf

			if spr->ObjectIndex1 > 0 then
				objst = @objects(spr->ObjectIndex1)->state(stateindex)
				VAddIm objst->force, springf
				'' relperp1 takes in to account distance from cm
				objst->torque += VDot(relperp1, springf)
			end if

			if spr->ObjectIndex2 > 0 then
				objst = @objects(spr->ObjectIndex2)->state(stateindex)
				VSubIm objst->force, springf
				'' relperp1 takes in to account distance from cm
				objst->torque -= VDot(relperp2, springf)
			end if

		end if
	next i

end sub

'' ==========================================================================

sub integrate_positions( byval dt as real, byval state1 as integer, byval state2 as integer)
	dim i as integer, t as vector
	dim as object_t ptr obj
	dim as objectstate_t ptr objst1, objst2, objst0
	for i = 1 to NUM_OBJECTS
		obj = objects(i)
		objst1 = @obj->state(state1)
		objst2 = @obj->state(state2)

		if obj->colcount = 0 then
			VMul t, dt, objst1->v
			VAdd objst2->p, objst1->p, t
			objst2->angp = objst1->angp + dt * objst1->angv

			VMul t, dt / obj->mass, objst1->force
			VAdd objst2->v, objst1->v, t
			objst2->angv = objst1->angv + dt / obj->moi * objst1->torque
		else
			objst2->p = objst1->p
			objst2->angp = objst1->angp
			objst2->v = objst1->v
			objst2->angv = objst1->angv
		end if


	next i
end sub

		
'' ==========================================================================
'' Collision Testing
'' ==========================================================================

'' determine is a point is near an object by check it's boundary box (+ EPSILON)

function IsPointNearObject(byval stateindex as integer, byval objectindex as integer, v as vector) as integer

	dim as integer j
	dim as shape_t ptr shp
	dim as objectstate_t ptr objst
	dim as polygon_t ptr ply
	dim as line_t f
	dim as real d
	dim relpnt as vector

	function = False

	shp = objects(objectindex)->shp
	objst = @objects(objectindex)->state(stateindex)

	'' Map to the shape's coordinates
	MapWorldPointToShape relpnt, objst, v

	ply = GetBoxPoly(shp)
	for j = 0 to ply->npoints - 1
		f = ply->faces[j]
		d = VDot(relpnt, f.n) + f.d
		if d >= EDGE_EPSILON then
			exit function
		end if
	next j

	function = True

end function


'' determine if a point is inside an object 
function IsPointInObject(byval stateindex as integer, byval objectindex as integer, v as vector, relpnt as vector) as integer
	dim as integer i, j, ninside, nply
	dim as shape_t ptr shp
	dim as objectstate_t ptr objst
	dim as polygon_t ptr ply
	dim as line_t f
	dim as real d
	IsPointInObject = False

	shp = objects(objectindex)->shp
	objst = @objects(objectindex)->state(stateindex)

	'' Map to the shape's coordinates
	MapWorldPointToShape relpnt, objst, v

	'' Test the object's box first
	ply = GetBoxPoly(shp)
	for j = 0 to ply->npoints - 1
		f = ply->faces[j]
		d = VDot(relpnt, f.n) + f.d
		if d >= 0 then
			exit function
		end if
	next j

	'' Test each polygon in the object
	nply = GetPolyCount(shp)
	for i = 0 to nply - 1
		ply = GetPoly(shp, i)
		ninside = 0
		for j = 0 to ply->npoints - 1
			f = ply->faces[j]
			d = VDot(relpnt, f.n) + f.d
			if d < 0 then
				ninside += 1
			end if
		next j
		if ninside = ply->npoints then
			function = True
			exit function
		end if
	next i
end function

'' given a point in the world, determine if it is inside an object
'' return a point relative to the obects coordinates / rotation
function FindObjectFromPoint(byval stateindex as integer, v as vector, relpnt as vector) as integer
	dim i as integer
	function = 0 '' none found
	for i = 1 to NUM_OBJECTS
		if IsPointInObject(stateindex, i, v, relpnt) = True then
			function = i
			exit for
		end if
	next i
end function

'' Determine if collision between point/velocity and an object
'' stateindex is the state to test
'' objectindex is the object to test
'' colpnt is a point in the world we are testing
'' colvel is the velocity that the point is moving
function FindPointToObjectCollision(collision as collision_t, byval stateindex as integer, byval objectindex as integer, colpnt as vector, colvel1 as vector) as integer
	dim as integer ret, i, j, nhit, ndeep, nedge, tmp_polyindex, tmp_faceindex, nply2
	dim as object_t ptr obj2
	dim as shape_t ptr shp2
	dim as objectstate_t ptr objst2
	dim as polygon_t ptr ply2
	dim as vector relvel, colcmp2, colprp2, colvel2
	dim as vector v, tmp_colnrm
	dim as line_t f
	dim as real d, relveldot, tmp_relveldot

	ret = CollisionState_None
	collision.state = ret
	function = ret

	'' Box test
	if IsPointNearObject(stateindex, objectindex, colpnt) = False then
		exit function
	end if

	obj2 = objects(objectindex)
	shp2 = obj2->shp
	objst2 = @obj2->state(stateindex)

	'' Calculate rel position of point wrt object #2
	VSub colcmp2, colpnt, objst2->p
	'' Calculate perp of object cm to point 
	VPerp colprp2, colcmp2

	'' Calculate inst. velocity at the point for object #2
	VMul   colvel2, objst2->angv, colprp2
	VAddIm colvel2, objst2->v

	'' Calculate Relative Velocity
	VSub relvel, colvel1, colvel2

	'' loop through all polygons
	nply2 = GetPolyCount(shp2)
	for i = 0 to nply2 - 1
		'' Find out if the point is inside a polygon
		ply2 = GetPoly(shp2, i)
		if (ply2->polyflags and BSP_SOLID) <> 0 then

			'' iterate each edge of the poly and test the side.
			nhit = 0
			ndeep = 0
			nedge = 0
			tmp_relveldot = 0

			for j = 0 to ply2->npoints - 1

				v = ply2->faces[j].n
				f.n.x = v.x * objst2->cs - v.y * objst2->sn
				f.n.y = v.x * objst2->sn + v.y * objst2->cs
				f.d = ply2->faces[j].d
				'' Behind face? How far?
				d = VDot(colcmp2, f.n) + f.d
				if d < 0 then
					nhit += 1
					ndeep += 1
				elseif d < EDGE_EPSILON then
					nhit += 1
					nedge += 1
					relveldot = VDot(f.n, relvel)						
					'' Moving towards object?
					if (relveldot < 0) then
						'' First contact edge found?
						if nedge = 1 then
							tmp_relveldot = relveldot
							''tmp_colnrm = ply2->axials[j].n '' f.n
							tmp_colnrm = f.n
							tmp_polyindex = i
							tmp_faceindex = j
						else
							'' point to near point contact
							'' FIXME: this really does not work that well (at all)
							'' shapes with sharp corners give it problems
							VAddIm tmp_colnrm, f.n
							''VUnitIm tmp_colnrm
						end if
					end if
				else
					exit for
				end if
			next j
			
			'' FIXME: handle inside edges - the following code will
			'' ignore a collision that is near the face of a polygon
			'' but is actually deep inside joined polygons

			'' if ndeep = ply->npoints - ply->numinsideedges then '' maybe ??
			if ndeep = ply2->npoints then
				collision.state = CollisionState_TooFar
				if collision.state > ret then
					ret = collision.state
				end if
				collision.normal = tmp_colnrm
				collision.objectindex2 = objectindex
				Add_Collision collision
				obj2->colcount += 1
			end if

			if nhit = ply2->npoints and tmp_relveldot < 0 then
				collision.state = CollisionState_Yes
				if collision.state > ret then
					ret = collision.state
				end if
				VUnitIm tmp_colnrm
				collision.normal = tmp_colnrm
				if collision.objectindex1 = -1 then
					VNegIm collision.normal
				end if
				collision.objectindex2 = objectindex
				Add_Collision collision
				obj2->colcount += 1
			end if

		end if
	next i
	function = ret
end function

'' determine if collision between point velocity and a wall
function FindPointToWorldCollision(collision as collision_t, colpnt as vector, colvel as vector) as integer
	dim as integer result
	dim as real d, relvel
	dim as vector n
	collision.state = CollisionState_None
	result = FindWorldCollision(colpnt, d, n)
	if result < 0 then
		collision.state = CollisionState_TooFar
		collision.normal = n
		Add_Collision collision
	elseif result = 0 then
		relvel = VDot(n, colvel)						
		if (relvel < 0) then
			collision.state = CollisionState_Yes
			collision.normal = n
			Add_Collision collision
		end if
	end if
	function = collision.state
end function

'' Determine if there are any collisions
sub FindCollisions(byval stateindex as integer) 
	dim as integer i, j, k, result
	dim as object_t ptr obj
	dim as shape_t ptr shp
	dim as objectstate_t ptr objst
	dim as polygon_t ptr ply
	dim as collision_t collision
		
	dim as vector colpnt, colcmp, colprp, colvel
	

	'' TODO: if something goes wrong with the simulation and an object ends up
	''       on the 'outside' of the world, FindPointToObjectCollision does not
	''       recognize that and returns a false positive.  i.e. the object is
	''       outside the world and moving back in but a collision with a world
	''       point prevents the object from getting back in.
	VSet colvel, 0, 0
	for i = 1 to NUM_WORLD_POINTS
		if (worldpointflags(i) and BSP_SOLID) <> 0 then
			colpnt = worldpoints(i)
			for j = 1 to NUM_OBJECTS
				collision.kind = CollisionType_WorldToObject
				collision.objectindex1 = -1
				collision.pointindex1 = i
				result = FindPointToObjectCollision(collision, stateindex, j, colpnt, colvel)
				if (result <> CollisionState_None) then
					objects(j)->colcount += 1
					objects(j)->colobject += 1
				end if
			next j
		end if
	next i

	'' Loop through all objects
	for i = 1 to NUM_OBJECTS
		obj = objects(i)
		shp = obj->shp
		objst = @objects(i)->state(stateindex)

		ply = GetOutlinePoly(shp)
		for j = 0 to ply->npoints - 1

			'' Calculate perp of object cm to point
			colpnt = objst->wpoints[ply->points[j]]
			VSub colcmp, colpnt, objst->p
			VPerp colprp, colcmp

			'' Calculate inst. velocity at the point
			VMul   colvel, objst->angv, colprp
			VAddIm colvel, objst->v

			collision.kind = CollisionType_ObjectToWorld
			collision.objectindex1 = i
			collision.pointindex1 = j
			result = FindPointToWorldCollision(collision, colpnt, colvel)
			if (result <> CollisionState_None) then
				obj->colcount += 1
				obj->colworld += 1
			end if

			for k = 1 to NUM_OBJECTS
				if k <> i then
					collision.kind = CollisionType_ObjectToObject
					collision.objectindex1 = i
					collision.pointindex1 = j
					result = FindPointToObjectCollision(collision, stateindex, k, colpnt, colvel)
					if (result <> CollisionState_None) then
						obj->colcount += 1
						obj->colobject += 1
						objects(k)->colcount += 1
						objects(k)->colobject += 1
					end if
				end if
			next k
		next j
	next i

end sub

'' ==========================================================================
'' Results of Collisions
'' ==========================================================================

sub ApplyCollision(byval stateIndex as integer, collision as collision_t)

	dim obj1 as object_t ptr, objst1 as objectstate_t ptr, ply1 as polygon_t ptr, shp1 as shape_t ptr
	dim obj2 as object_t ptr, objst2 as objectstate_t ptr
	dim as real perpdot1, perpdot2, ImpulseN, ImpulseD, Impulse, d, s
	dim as vector dv, t

	dim as vector colpnt, colcmp1, colprp1, colvel1, colcmp2, colprp2, colvel2, relvel

	Dbg_Print( "ApplyCollision( stateindex=" & str(stateindex) & " )" )
	Dbg_Print_Collision(collision)

	select case collision.kind
	case CollisionType_WorldToObject
		obj1 = objects(collision.objectindex2)
		shp1 = obj1->shp
		objst1 = @obj1->state(stateIndex)
		ply1 = GetOutlinePoly(shp1)

		colpnt = worldpoints(collision.pointindex1)

	case else
		obj1 = objects(collision.objectindex1)
		shp1 = obj1->shp
		objst1 = @obj1->state(stateIndex)
		ply1 = GetOutlinePoly(shp1)

		'' colpnt = Point of collision
		colpnt = objst1->wpoints[ply1->points[collision.pointindex1]]

	end select

	'Dbg_Print("  colpnt=" & VString(colpnt) )

	'' Calculate perp of object #1 cm to point 
	VSub colcmp1, colpnt, objst1->p
	VPerp colprp1, colcmp1

	'Dbg_Print("  p1=" & VString(objst1->p) )
	'Dbg_Print("  colcmp1=" & VString(colcmp1) )
	'Dbg_Print("  colprp1=" & VString(colprp1) )

	select case collision.kind
	case CollisionType_ObjectToObject
		if obj1->unresolved > 5 then
			VSet objst1->v, 0, 0
			objst1->angv = 0
		
			obj2 = objects(collision.objectindex2)
			objst2 = @obj2->state(stateIndex)
			VSet objst2->v, 0, 0
			objst2->angv = 0
		
			Dbg_Print("  *** v1=" & VString(objst1->v) & ", angv1=" & str(objst1->angv) )

			collision.state = -1
		else
			obj2 = objects(collision.objectindex2)
			objst2 = @obj2->state(stateIndex)

			'' Calculate perp of object #2 cm to point 
			VSub colcmp2, colpnt, objst2->p
			VPerp colprp2, colcmp2

			'Dbg_Print("  p2=" & VString(objst2->p) )
			'Dbg_Print("  colcmp2=" & VString(colcmp2) )
			'Dbg_Print("  colprp2=" & VString(colprp2) )

			'' Calculate inst. velocity at the point for object #1
			VMul   colvel1, objst1->angv, colprp1
			VAddIm colvel1, objst1->v

			Dbg_Print("  v1=" & VString(objst1->v) & ", angv1=" & str(objst1->angv) )
			Dbg_Print("  colvel1=" & VString(colvel1) )

			'' Calculate inst. velocity at the point for object #2
			VMul   colvel2, objst2->angv, colprp2
			VAddIm colvel2, objst2->v

			Dbg_Print("  v2=" & VString(objst2->v) )
			Dbg_Print("  angv2=" & str(objst2->angv) )
			Dbg_Print("  colvel2=" & VString(colvel2) )

			'' Calculate Relative Velocity
			VSub relvel, colvel1, colvel2
			Dbg_Print("  relvel=" & VString(relvel) )

			d = VDot(relvel,collision.normal)
			Dbg_Print("    d=" & str(d) )

			if d < 0 then
				collisionstate.unresolved += 1
				obj1->unresolved += 1

				'' Do a simple average on the two objects coefficent's of restitution
				ImpulseN = -(1.0 + (obj1->cor + obj2->cor)/2) * VDot(relvel,collision.normal)

				perpdot1 = VDot(colprp1,collision.normal)
				perpdot2 = VDot(colprp2,collision.normal)

				'' Collision normal is known to be unit
				ImpulseD  = (1.0/obj1->mass + 1.0/obj2->mass)
				ImpulseD += perpdot1 * perpdot1 / obj1->moi
				ImpulseD += perpdot2 * perpdot2 / obj2->moi

				Impulse = ImpulseN / ImpulseD

				'Dbg_Print("  ImpulseN=" & str(ImpulseN) )
				'Dbg_Print("  perpdot1=" & str(perpdot1) )
				'Dbg_Print("  perpdot2=" & str(perpdot2) )
				'Dbg_Print("  ImpulseD=" & str(ImpulseD) )
				Dbg_Print("  Impulse=" & str(Impulse) )

				'' Effect on object #1
				VMul dv, Impulse / obj1->mass, collision.normal
				VAddImMinimal objst1->v, dv
				''objst1->angv += Impulse / obj1->moi * perpdot1
				objst1->angv = AddMinimal(objst1->angv , Impulse / obj1->moi * perpdot1)

				Dbg_Print("  dv1=" & VString(dv) )
				Dbg_Print("  v1'=" & VString(objst1->v) )
				Dbg_Print("  angv1'=" & str(objst1->angv) )

				'' Effect on object #2 opposite of object #1
				VMul dv, Impulse / obj2->mass, collision.normal
				VSubImMinimal objst2->v, dv
				'' objst2->angv -= Impulse / obj2->moi * perpdot2
				objst2->angv = SubMinimal(objst2->angv, Impulse / obj2->moi * perpdot2)

				Dbg_Print("  dv2=" & VString(dv) )
				Dbg_Print("  v2'=" & VString(objst2->v) )
				Dbg_Print("  angv2'=" & str(objst2->angv) )
			else
				collision.state = -1
			end if
		end if

	case else

		'' Simplified case (walls don't move), relative velocity is colvel1
		'' TODO: define a coefficient of restitution for the wall (how springy is the wall?)
		if obj1->unresolved > 5 then
			VMulIm objst1->v, -obj1->cor
			objst1->angv *= -obj1->cor
			Dbg_Print("  *** v1=" & VString(objst1->v) & ", angv1=" & str(objst1->angv) )
			collision.state = -1
		else

			'' Calculate inst. velocity at the point for object #1
			VMul   colvel1, objst1->angv, colprp1
			VAddIm colvel1, objst1->v

			Dbg_Print("  v1=" & VString(objst1->v) & ", angv1=" & str(objst1->angv) )
			Dbg_Print("  colvel1=" & VString(colvel1) )

			d = VDot(colvel1, collision.normal)
			Dbg_Print("    d=" & str(d) )

			if d < 0 then
				collisionstate.unresolved += 1
				obj1->unresolved += 1
				ImpulseN = -(1.0 + obj1->cor) * VDot(colvel1,collision.normal)
				perpdot1 = VDot(colprp1,collision.normal)
				ImpulseD = 1.0/obj1->mass + perpdot1 * perpdot1 / obj1->moi
				Impulse = ImpulseN / ImpulseD

				'Dbg_Print("  ImpulseN=" & str(ImpulseN) )
				'Dbg_Print("  perpdot1=" & str(perpdot1) )
				'Dbg_Print("  ImpulseD=" & str(ImpulseD) )
				Dbg_Print("  Impulse=" & str(Impulse) )

				VMul dv, Impulse / obj1->mass, collision.normal
				VAddImMinimal objst1->v, dv
				'' objst1->angv += Impulse / obj1->moi * perpdot1
				objst1->angv = AddMinimal(objst1->angv, Impulse / obj1->moi * perpdot1)

				Dbg_Print("  dv1=" & VString(dv) )
				Dbg_Print("  v1'=" & VString(objst1->v) )
				Dbg_Print("  angv1'=" & str(objst1->angv) )
			else
				collision.state = -1
			end if
		end if

	end select

end sub

sub ApplyCollisions( byval stateindex as integer)
	dim as integer i, j
	j = 0
	do
		collisionstate.unresolved = 0
		for i = 1 to collisionstate.count
			select case collisions(i).kind
			case CollisionType_ObjectToWorld, CollisionType_WorldToObject
				ApplyCollision(stateindex, collisions(i))
			end select
		next i
		for i = 1 to collisionstate.count
			if collisions(i).kind = CollisionType_ObjectToObject then
				ApplyCollision(stateindex, collisions(i))
			end if
		next i
		j += 1
	loop while collisionstate.unresolved > 0
end sub

'' ==========================================================================

'' Forward the simulation by a time step (dt)
function RenderFrame(byval dt as double ) as double
	dim as double t1, t2
	dim i as integer

	t1 = 0
	t2 = dt

	'' Fix time step to limit errors when dt is large
	if dt > 0.05 then
		dt = 0.05
	end if

	while (t1 < dt)

		'' Needed incase simulation locks
		if multikey(sc_escape) then
			exit while
		end if

		'' Dbg_Print(str(t2-t1) & " " & str(t1) & " " & str(t2))
		
		Dbg_Print ( "Time Step: " & str(t2- t1) )

		ComputeForces(STATE1)
		Integrate_Positions(t2-t1, STATE1, STATE2)
		RenderObjectsToWorld(STATE2)

		Reset_Collisions()
		FindCollisions(STATE2)

		if (collisionstate.state = CollisionState_TooFar) then
			t2 = (t1 + t2) / 2
		else

			'' Collision Resolution
			'' TODO: make this time dependant
			i = 0
			while (collisionstate.state <> CollisionState_None) and (i < 10)
				'' Needed incase simulation lock
				if multikey(sc_escape) then
					exit while
				end if

				ApplyCollisions(STATE2)

				'STATE1 xor= 1
				'STATE2 xor= 1
				'ComputeForces(STATE1)
				'Integrate_Positions(0, STATE1, STATE2)
				'RenderObjectsToWorld(STATE2)

				Reset_Collisions()
				FindCollisions(STATE2)
				FindCollisions(STATE1)
				i += 1
			wend

			'' Shock Propagation
			i = 0
			'while (collisionstate.state <> CollisionState_None) and (i < 5)
			'wend

			t1 = t2
			t2 = dt
			STATE1 xor= 1
			STATE2 xor= 1
			
		end if
		
	wend
	function = 0
end function

