'' ============================================================
'' 2D Rigid Body Dynamics Simulation Engine
''
'' Copyright (C) 2006 - Jeff Marshall (coder@execulink.com)
''
'' This software is provided 'as-is', without any express or implied
'' warranty. In no event will the authors be held liable for any damages
'' arising from the use of this software.
''	
'' Permission is granted to anyone to use this software for any purpose,
'' including commercial applications, and to alter it and redistribute it
'' freely, subject to the following restrictions:
''	
'' 1. The origin of this software must not be misrepresented; you must not
'' claim that you wrote the original software. If you use this software in
'' a product, an acknowledgment in the product documentation would be
'' appreciated but is not required.
''	
'' 2. Altered source versions must be plainly marked as such, and must not
'' be misrepresented as being the original software.
''	
'' 3. This notice may not be removed or altered from any source
'' distribution.
''
'' ============================================================

Option Explicit

#include "rigid_types.bi"
#include "vector.bi"
#include "rigid_world.bi"
#include "rigid_shapes.bi"
#include "rigid_engine.bi"
#include "rigid_scene.bi"
#include "rigid_bsp.bi"
#include "cons_log.bi"

dim Shared SCENE_TITLE as string
redim Shared SCENE_DESC(0 to 50) as string
dim Shared SCENE_DESC_COUNT as integer
dim Shared SCENE_DESC_TIMER as double

enum ResultEnum
	ERROR_NONE = 0
	INTERNAL_ERROR
	FILE_NOT_FOUND
	FILE_NOT_SPECIFIED
	BLOCKS_NESTED_TOO_DEEP
	END_BLOCK_UNDERFLOW
	INCLUDE_NESTED_TOO_DEEP
	INCLUDE_FILE_UNDERFLOW
	UNKNOWN_KEYWORD
	INVALID_END_BLOCK
	CIRCULAR_INCLUDE
	SHAPE_NOT_DEFINED
	INVALID_AREA_TYPE
end enum

enum BlockTypeEnum
	BT_None = 0
	BT_Ignore
	BT_Shape
	BT_Object
	BT_Description
	BT_Area
end enum

enum BlockSubTypeEnum
	BST_normal = 0
	BST_solid = 1
	BST_force = 2
end enum

Type BlockType
	btType as integer '' ENUM:BlockTypeEnum
	btSubType as integer '' ENUM:BlockSubTypeEnum
	btFlags as integer
	btForce as vector
	btPointCount as integer
	btFirstPoint as Vector
	btLastPoint as Vector
	btOrigin as Vector
End Type

Type FileType
	ftName as string
	ftLineNo as integer
	ftErrorNo as integer
	ftErrorMsg as string
End Type

const MAX_BLOCK_DEPTH = 20
Dim Shared BlockLevel as integer
Dim Shared BlockStack(0 to MAX_BLOCK_DEPTH - 1) as BlockType

const MAX_FILE_DEPTH = 20
Dim Shared FileLevel as integer
Dim Shared FileStack(0 to MAX_FILE_DEPTH - 1) as FileType

declare function LoadSceneFile(strFileName as string) as integer

'' ==========================================================================
'' INTERNAL
'' ==========================================================================

'' ftErrorNo ENUM:ResultEnum
'' return    ENUM:ResultEnum
private function SetErrorMsg(ftErrorNo as integer, ftErrorMsg as string) as integer
	if FileLevel > 0 then
		with FileStack(FileLevel - 1)
			.ftErrorNo = ftErrorNo
			select case ftErrorNo
			case FILE_NOT_FOUND
				.ftErrorMsg = "File not found, " & ftErrorMsg
			case FILE_NOT_SPECIFIED
				.ftErrorMsg = "File not specified " & ftErrorMsg
			case UNKNOWN_KEYWORD
				.ftErrorMsg = "Unrecognized keyword, " & ftErrorMsg
			case INVALID_END_BLOCK
				.ftErrorMsg = "Invalid end block, " & ftErrorMsg
			case CIRCULAR_INCLUDE
				.ftErrorMsg = "Circularly included file, " & ftErrorMsg
			case SHAPE_NOT_DEFINED
				.ftErrorMsg = "shape not defined, " & ftErrorMsg
			case INVALID_AREA_TYPE
				.ftErrorMsg = "Invalid area type, " & ftErrorMsg
			case else
				.ftErrorMsg = ftErrorMsg
			end select
		end with
		function = ftErrorNo
	else
		function = INTERNAL_ERROR
	end if
end function

'' btType ENUM:BlockTypeEnum
'' return ENUM:ResultEnum
private function PushBlock(byval btType as integer, byval btSubType as integer) as integer
	function = False
	if BlockLevel	< MAX_BLOCK_DEPTH then
		with BlockStack(BlockLevel)
			.btType = btType
			.btSubType = btSubType
			.btPointCount = False
			.btFlags = 0
			VSet .btForce, 0, 0
			VSet .btFirstPoint, 0, 0
			VSet .btLastPoint, 0, 0
			VSet .btOrigin, 0, 0
		end with
		BlockLevel += 1
		function = ERROR_NONE
	else
		function = SetErrorMsg( BLOCKS_NESTED_TOO_DEEP, "Blocks Nested too deep" )
	end if
end function

private function PopBlock() as integer
	function = False
	if BlockLevel > 0 then
		BlockLevel -= 1
		function = ERROR_NONE
	else
		SetErrorMsg END_BLOCK_UNDERFLOW, "end block underflow"
		function = END_BLOCK_UNDERFLOW
	end if
end function

private function PushFile(strFileName as string) as integer
	dim i as integer
	if FileLevel < MAX_FILE_DEPTH then
		for i = 0 to FileLevel - 1
			if lcase(FileStack(i).ftname) = lcase(strFileName) then
				function = SetErrorMsg( CIRCULAR_INCLUDE, strFileName )
				exit function
			end if
		next i
		with FileStack(FileLevel)
			.ftName = strFileName
			.ftLineNo = 0
			.ftErrorNo = 0
			.ftErrorMsg = ""
		end with
		FileLevel += 1
		function = ERROR_NONE
	else
		function = SetErrorMsg( INCLUDE_NESTED_TOO_DEEP, "Include Nested Too Deep" )
	end if
end function

private function PopFile() as integer
	if FileLevel > 0 then
		function = FileStack(FileLevel - 1).ftErrorNo
		FileLevel -= 1
	else
		function = INTERNAL_ERROR
	end if
end function

'' ==========================================================================
'' SCENE LOADING
'' ==========================================================================

private function stripquotes(x as string) as string
	function = x
	if len(x) > 1 then
		if (left(x,1) = "'" and right(x,1) = "'") or _
		  (left(x,1) = chr(34) and right(x,1) = chr(34)) then
		
			function = mid(x, 2, len(x) - 2)
		end if
	end if
end function

private sub ParseLine(x as string, w() as string, n as integer, s2n as string, m as integer=0)
	dim i as integer, j as integer
	n = 0
	i = 1
	s2n = ""
	if m = 0 then
		m = ubound(w)
	end if
	while n < m
		while mid(x, i ,1) = " " or mid(x, i ,1) = chr(9)
			i += 1
		wend
		j = instr(i, x, any ", ")
		if j = 0 then
			n = n + 1
			w(n) = stripquotes(trim(mid(x,i), any " " & chr(9)))
			exit while
		else
			if n = 1 then
				s2n = trim(mid(x,i), any " " & chr(9))
			end if
			n = n + 1
			w(n) = stripquotes(trim(mid(x,i,j-i), any " " & chr(9)))
			i = j + 1
		end if
	wend
end sub

private function processline(byval lineno as integer, s as string) as integer
	dim w(1 to 32) as string, s2n as string
	dim as integer n, m, i, j, ret, startidx, b
	static as shape_t ptr shp, shp2
	static ply as polygon_t ptr
	static obj as object_t ptr
	static as vector p1, p2, t

	ret = ERROR_NONE '' assume no error
	function = ret 
	ParseLine s, w(), n, s2n

	Select case  BlockStack(BlockLevel - 1).btType
	case BT_Ignore
		select case lcase(w(1))
		case "ignore"
			ret = PushBlock(BT_Ignore, BST_Normal)
		case "end"
			if lcase(w(2)) = "ignore" then
			  ret = PopBlock()
			end if
		end select

	case BT_Description
		if lcase(w(1)) = "end" and lcase(w(2)) = "description" then
			ret = PopBlock()
		else
			SCENE_DESC_COUNT += 1
			SCENE_DESC(SCENE_DESC_COUNT) = s
			Log_Print(s)
		end if

	case BT_Area
		select case lcase(w(1))
		case "direction"
			with BlockStack(BlockLevel - 1)
				VSet .btForce, val(w(2)), val(w(3))
			end with
		case "origin"
			with BlockStack(BlockLevel - 1)
				VSet .btOrigin, val(w(2)), val(w(3))
			end with
		case "point"
			with BlockStack(BlockLevel - 1)
				if .btPointCount = 0 then
					VSet .btFirstPoint, val(w(2)), val(w(3))
					VAddIm .btFirstPoint, .btOrigin
					.btLastPoint = .btFirstPoint
					.btPointCount = 1
				else
					VSet p1, val(w(2)), val(w(3))
					VAddIm p1, .btOrigin
					Add_WallByPoints(.btLastPoint, p1, .btFlags, .btForce)
					.btLastPoint = p1
					.btPointCount += 1
				endif
			end with
		case "end"
			if lcase(w(2)) = "area" then
				with BlockStack(BlockLevel - 1)
					if .btPointCount > 2 then
						Add_WallByPoints(.btLastPoint, .btFirstPoint, .btFlags, .btForce)
						.btPointCount += 1
					end if
				end with
				ret = PopBlock()
				Dbg_Print("Area Added")
			else
				ret = SetErrorMsg( INVALID_END_BLOCK, w(1) & " " & w(2) )
			end if
		case else
			ret = SetErrorMsg( UNKNOWN_KEYWORD, w(1) )
		end select

	case BT_Shape
		select case lcase(w(1))
		case "color"
			SetCurrentColor val(w(2)), val(w(3)), val(w(4))
		case "point"
			if lcase(w(4)) = "nohit" then
				add_shape_point shp, val(w(2)), val(w(3)), BSP_NOTSOLID
			else
				add_shape_point shp, val(w(2)), val(w(3)), BSP_SOLID
			end if
		case "polygon"
			ply = create_polygon()
			ply->polyflags = BSP_SOLID
			for i = 2 to n
				if lcase(w(i)) = "nohit" then
					ply->polyflags = BSP_NOTSOLID
				else
					if i < n then
						if lcase(w(i+1)) = "i" then
						  add_polygon_point ply, val(w(i)), BSP_INVISIBLE
							i = i + 1
						else
							add_polygon_point ply, val(w(i)), BSP_SOLID
						end if
					else
						add_polygon_point ply, val(w(i)), BSP_SOLID
					end if
				end if
			next i
			Add_Shape_Polygon(shp,ply)
		case "insert"
			shp2 = Find_Shape(w(2))
			if shp2 then
				'' Copy shape 2 in to shape 1
				startidx = shp->npoints
				for i = 0 to shp2->npoints - 1
					add_shape_point shp, shp2->lpoints[i].x, shp2->lpoints[i].y, shp2->lpointflags[i]
				next i

				for i = 0 to GetPolyCount(shp2) - 1
					ply = create_polygon()
					ply->polyflags = GetPoly(shp2,i)->polyflags
					for j = 0 to GetPoly(shp2,i)->npoints - 1
						add_polygon_point ply, GetPoly(shp2,i)->points[j] + startidx, GetPoly(shp2,i)->pointflags[j]
					next j
					Add_Shape_Polygon(shp,ply)
				next i
			else
				ret = SetErrorMsg( SHAPE_NOT_DEFINED, w(2) )
			end if
		case "origin"
			VSet shp->origin, val(w(2)), val(w(3))
		case "centerofmass"
			VSet shp->cm, val(w(2)), val(w(3))
		case "end"
			if lcase(w(2)) = "shape" then
				Add_Shape(shp)
				ret = PopBlock()
				Dbg_Print("Added Shape '" & Trim(shp->name) & "'")
			else
				ret = SetErrorMsg( INVALID_END_BLOCK, w(1) & " " & w(2) )
			end if
		case else
			ret = SetErrorMsg( UNKNOWN_KEYWORD, w(1) )
		end select

	case BT_Object
		select case lcase(w(1))
		case "mass"	
			obj->mass = val(w(2))
		case "momentofinertia", "moi"
			obj->moi = val(w(2))
		case "coefficientofrestitution", "cor"
			obj->cor = val(w(2))
		case "position", "p"
			VSet obj->state(0).p, val(w(2)), val(w(3))
		case "angle"
			obj->state(0).angp = val(w(2))
		case "velocity", "v"
			VSet obj->state(0).v, val(w(2)), val(w(3))
		case "angularvelocity", "angv"
			obj->state(0).angv = val(w(2))
		case "end"
			if lcase(w(2)) = "object" then
				Add_object(obj)
				ret = PopBlock()
				Dbg_Print("Added Object")
			else
				ret = SetErrorMsg( INVALID_END_BLOCK, w(1) & " " & w(2) )
			end if
		case else
			ret = SetErrorMsg( UNKNOWN_KEYWORD, w(1) )
		end select

	case BT_None

		select case lcase(w(1))
		case "gravity"
			VSet AGRAVITY, val(w(2)), val(w(3))
		case "default_spring_ks"
			DEFAULT_SPRING_KS = val(w(2))
		case "default_spring_kd"
			DEFAULT_SPRING_KD = val(w(2))
		case "default_coefficientofrestitution", "default_cor"
			DEFAULT_COR = val(w(2))
		case "wall"
			VSet p1, val(w(2)), val(w(3))
			VSet p2, val(w(4)), val(w(5))
			VSet t, 0, 0			
			Add_WallByPoints(p1, p2, BSP_SOLID, t) '' FB 0.15 COmpatible
			'' Add_WallByPoints(p1, p2, BSP_SOLID, Type(0,0))
		case "area"
			select case lcase(w(2))
			case "solid"
				ret = PushBlock(BT_Area, BST_Solid)
				BlockStack(BlockLevel - 1).btFlags = BSP_SOLID
			case "force"
				ret = PushBlock(BT_Area, BST_Force)
				BlockStack(BlockLevel - 1).btFlags = BSP_NOTSOLID or BSP_FORCE
			case else
				ret = SetErrorMsg( INVALID_AREA_TYPE, w(2) )
			end select
		case "title"
			SCENE_TITLE = s2n
			Log_Print("Title: " & SCENE_TITLE)
		case "shape"
			shp = create_shape(w(2))
			ret = PushBlock(BT_Shape, BST_Normal)
		case "object"
			shp = Find_Shape(w(2))
			if (shp) then
				obj = Create_object()
				obj->shp = Find_Shape(w(2))
				ret = PushBlock(BT_Object, BST_Normal)
			else
				ret = SetErrorMsg( SHAPE_NOT_DEFINED, w(2) )
			end if
		case "description"
			SCENE_DESC_TIMER = val(w(2))
			ret = PushBlock(BT_Description, BST_Normal)
		case "ignore"
			ret = PushBlock(BT_Ignore, BST_Normal)
		case "include"
			ret = LoadSceneFile(w(2))
		case else
			ret = SetErrorMsg( UNKNOWN_KEYWORD, w(1) )
		end select
	end select

	function = ret

end function

private function LoadSceneFile(strFileName as string) as integer
	dim as string x, s
	dim as integer h, ret, n = 0
	If FileLevel > 1 then
		Log_Print("Including '" & strFileName & "'")
	else
		Log_Print("Loading '" & strFileName & "'")
	end if
	if Trim(strFileName) = "" then
		function = SetErrorMsg(FILE_NOT_SPECIFIED,"")
	elseif dir(strFileName) = "" then
		function = SetErrorMsg(FILE_NOT_FOUND, strFileName)
	else
		ret = PushFile(strFileName)
		if ret = ERROR_NONE then
			h = freefile
			open strFileName for input as #h
			while not eof(h)
				line input #h, x
				FileStack(FileLevel-1).ftLineNo += 1
				n += 1
				s = trim(x, any " " & chr(9))
				if left(s,1) <> "'" and left(s,1) <> ";" and left(s,1) <> "#" and s > "" then
					ret = processline (n, s)
					if ret <> 0 then	
						exit while
					end if
				end if
			wend
			close #h
		end if
		if ret = ERROR_NONE then
			ret = PopFile()
		end if
		function = ret
	end if
end function

private sub ReportSceneErrors()
	dim i as integer
	dim msg as string 
	if FileLevel > 0 then
		if FileStack(FileLevel - 1).ftErrorNo <> ERROR_NONE then	
			for i = 0 to FileLevel - 1
				msg = ""
				with FileStack(i)
					if .ftName > "" then
						msg = msg & .ftName
						if .ftLineNo > 0 then
							msg = msg & "(" & str(.ftLineNo) & ")"
						else
							msg = msg & "()"
						end if
					end if
					if .ftErrorNo <> ERROR_NONE then
						if msg > "" then msg = msg & " : "
						msg = msg & "Error " & str(.ftErrorNo) & " : "
						if .ftErrorMsg > "" then
							msg = msg & .ftErrorMsg
						end if
					end if
					if msg > "" then
						Log_Print(space(i*2) + msg)
					end if
				end with
			next i
		end if	
	end if

end sub

function LoadScene(strFileName as string) as integer
	dim ret as integer

	'' Setup defaults
	SCENE_TITLE = strFileName
	SCENE_DESC_COUNT = 0
	SCENE_DESC_TIMER = 0
	SetCurrentColor 255, 255, 255

	default_spring_ks = 200
	default_spring_kd = 20
	default_cor = 0.9

	FileLevel = 0
	PushFile("")

	BlockLevel = 0
	PushBlock BT_None, BST_Normal

	ret = LoadSceneFile(strFileName)

	PopBlock()

	ReportSceneErrors

	while FileLevel > 0
		PopFile()
	wend

	function = ret

end function
