'' ============================================================
'' 2D Rigid Body Dynamics Simulation Engine
''
'' Copyright (C) 2006 - Jeff Marshall (coder@execulink.com)
''
'' This software is provided 'as-is', without any express or implied
'' warranty. In no event will the authors be held liable for any damages
'' arising from the use of this software.
''	
'' Permission is granted to anyone to use this software for any purpose,
'' including commercial applications, and to alter it and redistribute it
'' freely, subject to the following restrictions:
''	
'' 1. The origin of this software must not be misrepresented; you must not
'' claim that you wrote the original software. If you use this software in
'' a product, an acknowledgment in the product documentation would be
'' appreciated but is not required.
''	
'' 2. Altered source versions must be plainly marked as such, and must not
'' be misrepresented as being the original software.
''	
'' 3. This notice may not be removed or altered from any source
'' distribution.
''
'' ============================================================

Option Explicit

#include "rigid_types.bi"
#include "rigid_shapes.bi"
#include "rigid_bsp.bi"

const PLIST_OUTLINE = 0
const PLIST_SHAPE = 1

const PLIST_OUTLINE_POINTS = 0
const PLIST_OUTLINE_BOX = 1

dim shared MAX_SHAPES as integer = 16
dim shared NUM_SHAPES as integer = 0
redim shared shapes(1 to MAX_SHAPES) as shape_t ptr
dim shared CURRENT_COLOR = rgb(255,255,255)

'' ==========================================================================
'' POLYGONS and POLYGON LISTS
'' ==========================================================================

'' Create a polygon (allocate space for it)
function Create_Polygon() as polygon_t ptr
	dim ply as polygon_t ptr
	ply = callocate(sizeof(polygon_t))
	if (ply) then
		ply->npoints = 0
		ply->points = 0 '' NULL
		ply->faces = 0 '' NULL
		ply->axials = 0 '' NULL
		ply->pointflags = 0 '' NULL
		ply->clr = CURRENT_COLOR
	end if
	function = ply
end function

'' add a point to a polygon's list of points as an index to the shape's list of points
'' return an index to the index in the list of points
function Add_Polygon_Point(byval ply as polygon_t ptr, byval pntidx as integer, byval flags as integer) as integer
	dim as integer ptr tmp1, tmp2
	tmp1 = reallocate(ply->points, (ply->npoints + 1) * sizeof(integer))
	tmp2 = reallocate(ply->pointflags, (ply->npoints + 1) * sizeof(integer))
	function = -1
	if (tmp1) <> 0 and (tmp2) <> 0 then
		ply->points = tmp1
		ply->pointflags = tmp2
		ply->points[ply->npoints] = pntidx
		ply->pointflags[ply->npoints] = flags
		function = ply->npoints
		ply->npoints += 1
	end if
end function

'' Destroy a polygon
sub Destroy_Polygon(byval ply as polygon_t ptr)
	if ply then
		if ply->pointflags then
			deallocate ply->pointflags
			ply->pointflags = 0 '' NULL
		end if
		if ply->points then 
			deallocate ply->points
			ply->points = 0 '' NULL
		end if
		if ply->faces then 
			deallocate ply->faces
			ply->faces = 0 '' NULL
		end if
		if ply->axials then 
			deallocate ply->axials
			ply->axials = 0 '' NULL
		end if
		deallocate ply
	end if
end sub

sub ComputePolyFaces(byval shp as shape_t ptr, byval ply as polygon_t ptr)

	'' Calculate edge normals and offset from center
	ply->faces = reallocate(ply->faces, sizeof(line_t) * ply->npoints)
	dim as vector a, b, ab
	dim as integer i
	for i = 0 to ply->npoints -1
		a = shp->lpoints[ply->points[i]]
		b = shp->lpoints[ply->points[(i+1) mod ply->npoints]]
		vSubIm a, shp->cm
		vSubIm b, shp->cm
		vSub ab, b, a
		VUnitIm ab
		ply->faces[i].n.x = -ab.y
		ply->faces[i].n.y = ab.x
		ply->faces[i].d = -VDot(a, ply->faces[i].n)
	next i
	VSet shp->cm, 0, 0
	'' FIXME: find common edges shared by joined polygons and mark the edge as inside
	'' Needed for proper collision detection

end sub

sub ComputePolyAxials(byval shp as shape_t ptr, byval ply as polygon_t ptr)
	'' Calculate point axials - assumes faces are already calculated
	ply->axials = reallocate(ply->axials, sizeof(line_t) * ply->npoints)
	dim as integer i
	for i = 0 to ply->npoints -1
		ply->axials[i].n = ply->faces[i].n
		VAddIm ply->axials[i].n, ply->faces[(i+ply->npoints-1) mod ply->npoints].n
		VUnitIm ply->axials[i].n
		ply->axials[i].d = -VDot(shp->lpoints[ply->points[i]], ply->axials[i].n)
	next i
	VSet shp->cm, 0, 0
end sub

'' Add a polygon to a list of polygons
function Add_Polygon(plylist as polygon_list_t ptr, byval ply as polygon_t ptr) as integer
	dim tmp as polygon_t ptr ptr 
	tmp = reallocate(plylist->polygons, (plylist->npolygons + 1) * sizeof(polygon_t ptr))
	function = 0
	if (tmp) then
		plylist->polygons = tmp
		plylist->polygons[plylist->npolygons] = ply
		function = plylist->npolygons
		plylist->npolygons += 1
	end if
end function

sub Destroy_Polygon_List(plylist as polygon_list_t ptr)
	dim i as integer
	if (plylist) then
		if (plylist->polygons) then
			for i = 0 to plylist->npolygons - 1
				if plylist->polygons[i] then
					Destroy_Polygon plylist->polygons[i]
					plylist->polygons[i] = 0 '' NULL
				end if
			next i
			deallocate plylist->polygons
			plylist->polygons = 0 '' NULL
		end if
	end if
end sub

'' ==========================================================================
'' SHAPES
'' ==========================================================================

'' Create a shape (allocate space for it)
function Create_Shape(strName as string) as shape_t ptr
	dim shp as shape_t ptr
	dim ply as polygon_t ptr
	shp = callocate(sizeof(shape_t))
	if (shp) then
		shp->name = strName
		VSet shp->origin,0,0
		VSet shp->cm,0,0
		ply = Create_Polygon()      '' create a polygon to store boundary points
		Add_Polygon(@shp->plists(PLIST_OUTLINE), ply)
		ply = Create_Polygon()      '' create a polygon to store boundary box
		Add_Polygon(@shp->plists(PLIST_OUTLINE), ply)
	end if
	function = shp
end function

'' Find a shape by case-insensitive name
function Find_Shape(strName as string) as shape_t ptr
	dim i as integer
	function = 0 '' NULL
	for i = 1 to NUM_SHAPES
		if lcase(strName) = rtrim(lcase(shapes(i)->name)) then
			function = shapes(i)
			exit for
		end if
	next i
end function

'' Add a point to a shape's list of points
'' Returns a zero based index of the vector in the point list
function Add_Shape_Point(byval shp as shape_t ptr, byval x as real, byval y as real, byval flags as integer) as integer
	dim tmp1 as vector ptr 
	dim tmp2 as integer ptr
	tmp1 = reallocate(shp->lpoints, (shp->npoints + 1) * sizeof(vector))
	tmp2 = reallocate(shp->lpointflags, (shp->npoints + 1) * sizeof(integer))
	
	function = 0
	if (tmp1) <> 0 and (tmp2 <> 0) then
		shp->lpoints = tmp1
		shp->lpointflags = tmp2
		shp->lpoints[shp->npoints].x = x + shp->origin.x
		shp->lpoints[shp->npoints].y = y + shp->origin.y
		shp->lpointflags[shp->npoints] = flags
		function = shp->npoints
		shp->npoints += 1
	end if
end function

private sub SetOutLinePoly(byval shp as shape_t ptr)
	dim i as integer
	dim ply as polygon_t ptr

	ply = GetOutlinePoly(shp)
	Destroy_Polygon(ply)

	ply = Create_Polygon()
	for i = 0 to shp->npoints - 1	
		if (shp->lpointflags[i] and BSP_SOLID) <> 0 then
			Add_Polygon_Point(ply, i, BSP_SOLID)
		end if
	next i

	'' This plist was already created in Create_Shape()
	shp->plists(PLIST_OUTLINE).polygons[PLIST_OUTLINE_POINTS] = ply
end sub

private sub SetBoxPoly(byval shp as shape_t ptr)
	dim i as integer, icount
	dim ply as polygon_t ptr
	dim as vector a, b, t

	icount = 0
	for i = 0 to shp->npoints - 1
		if (shp->lpointflags[i] and BSP_SOLID) <> 0 then	
			if icount = False then
				icount += 1
				a = shp->lpoints[i]
				b = a
			else
				if a.x < shp->lpoints[i].x then a.x = shp->lpoints[i].x
				if b.x > shp->lpoints[i].x then b.x = shp->lpoints[i].x
				if a.y < shp->lpoints[i].y then a.y = shp->lpoints[i].y
				if b.y > shp->lpoints[i].y then b.y = shp->lpoints[i].y
			end if
		end if		
	next i

	if icount > 0 then
		ply = GetBoxPoly(shp)
		Destroy_Polygon(ply)
		ply = Create_Polygon()
		i = Add_Shape_Point(shp, a.x, a.y, BSP_NOTSOLID)
		Add_Polygon_Point(ply, i, BSP_SOLID)
		i = Add_Shape_Point(shp, a.x, b.y, BSP_NOTSOLID)
		Add_Polygon_Point(ply, i, BSP_SOLID)
		i = Add_Shape_Point(shp, b.x, b.y, BSP_NOTSOLID)
		Add_Polygon_Point(ply, i, BSP_SOLID)
		i = Add_Shape_Point(shp, b.x, a.y, BSP_NOTSOLID)
		Add_Polygon_Point(ply, i, BSP_SOLID)
		shp->plists(PLIST_OUTLINE).polygons[PLIST_OUTLINE_BOX] = ply
		ComputePolyFaces shp, ply
		ComputePolyAxials shp, ply
	end if

end sub

'' Add a polygon to a shape's list of polygons
'' returns the zero based index of the polygon in the list
function Add_Shape_Polygon(byval shp as shape_t ptr, byval ply as polygon_t ptr) as integer
	ComputePolyFaces shp, ply
	ComputePolyAxials shp, ply
	SetOutlinePoly shp
	SetBoxPoly shp
	function = Add_Polygon(@shp->plists(PLIST_SHAPE), ply)
end function

'' add a shape to the global list of shapes
function Add_Shape(byval shp as shape_t ptr) as integer
	function = False
	if NUM_SHAPES = MAX_SHAPES then
		MAX_SHAPES = MAX_SHAPES * 2
		redim preserve shapes(1 to MAX_SHAPES) as shape_t ptr
	end if
	NUM_SHAPES = NUM_SHAPES + 1
	
	shapes(NUM_SHAPES) = shp
	function  = NUM_SHAPES
end function

'' set the current color
sub SetCurrentColor(r as integer, g as integer, b as integer)
	CURRENT_COLOR = rgb(r,g,b)
end sub


sub Destroy_Shape(byval shp as shape_t ptr)
	if (shp) then
		Destroy_Polygon_List @shp->plists(PLIST_OUTLINE)
		Destroy_Polygon_List @shp->plists(PLIST_SHAPE)
		if shp->lpoints then 
			Deallocate shp->lpoints
			shp->lpoints = 0 '' NULL
		end if
		if shp->lpointflags then
			Deallocate shp->lpointflags
			shp->lpointflags = 0 '' NULL
		end if
		deallocate shp
	end if
end sub

'' ==========================================================================
'' SHAPE HELPERS
'' ==========================================================================

'' dellocate all shapes (and associated polygons / points)
sub Destroy_All_Shapes()
	dim as integer i
	for i = 1 to NUM_SHAPES
		Destroy_Shape shapes(i)
		shapes(i) = 0 '' NULL
	next i
	NUM_SHAPES = 0
end Sub

function GetOutlinePoly(byval shp as shape_t ptr) as polygon_t ptr
	function = shp->plists(PLIST_OUTLINE).polygons[PLIST_OUTLINE_POINTS]
end function

function GetBoxPoly(byval shp as shape_t ptr) as polygon_t ptr
	function = shp->plists(PLIST_OUTLINE).polygons[PLIST_OUTLINE_BOX]
end function

function GetPolyCount(byval shp as shape_t ptr) as integer
	function = shp->plists(PLIST_SHAPE).npolygons
end function

function GetPoly(byval shp as shape_t ptr, index as integer) as polygon_t ptr
	function = shp->plists(PLIST_SHAPE).polygons[index]
end function
