'' ============================================================
'' 2D Rigid Body Dynamics Simulation Engine
''
'' Copyright (C) 2006 - Jeff Marshall (coder@execulink.com)
''
'' This software is provided 'as-is', without any express or implied
'' warranty. In no event will the authors be held liable for any damages
'' arising from the use of this software.
''	
'' Permission is granted to anyone to use this software for any purpose,
'' including commercial applications, and to alter it and redistribute it
'' freely, subject to the following restrictions:
''	
'' 1. The origin of this software must not be misrepresented; you must not
'' claim that you wrote the original software. If you use this software in
'' a product, an acknowledgment in the product documentation would be
'' appreciated but is not required.
''	
'' 2. Altered source versions must be plainly marked as such, and must not
'' be misrepresented as being the original software.
''	
'' 3. This notice may not be removed or altered from any source
'' distribution.
''
'' ============================================================

Option Explicit

#include "rigid_types.bi"
#include "rigid_bsp.bi"
#include "rigid_world.bi"

dim shared MAX_WALLS as integer = 16
dim shared NUM_WALLS as integer = 0
redim shared walls(1 to MAX_WALLS) as wall_t ptr

dim shared MAX_WORLD_POINTS as integer = 16
dim shared NUM_WORLD_POINTS as integer = 0
redim shared worldpoints(1 to MAX_WORLD_POINTS) as vector
redim shared worldpointflags(1 to MAX_WORLD_POINTS) as integer

dim shared world_bsp as bsp_tree_t ptr

'' ==========================================================================
'' POINTS
'' ==========================================================================

'' Add a point to the wpoints and return an index
function Add_World_Point(a as vector, byval flags as integer) as integer
	dim i as integer
	for i = 1 to NUM_WORLD_POINTS
		if abs(a.x - worldpoints(i).x) < POINT_EPSILON then
			if abs(a.y - worldpoints(i).y) < POINT_EPSILON then
				exit for
			end if
		end if
	next i
	if i > NUM_WORLD_POINTS then
		NUM_WORLD_POINTS += 1
		if NUM_WORLD_POINTS > MAX_WORLD_POINTS then
			MAX_WORLD_POINTS *= 2
			redim preserve worldpoints(1 to MAX_WORLD_POINTS) as vector
			redim preserve worldpointflags(1 to MAX_WORLD_POINTS) as integer
		end if
		i = NUM_WORLD_POINTS
		worldpointflags(i) = 0
	end if
	worldpoints(i) = a
	worldpointflags(i) or= flags
	function = i
end function

'' ==========================================================================
'' WALL CREATION
'' ==========================================================================

'' Create a new wall (allocate space for it)
function Create_wall() as wall_t ptr
	dim wall as wall_t ptr
	wall = callocate(sizeof(wall_t))
	function = wall
end function

'' Add a wall to the list of walls and return a zerobased index
function Add_Wall(byval wall as wall_t ptr) as integer
	function = False
	if NUM_WALLS = MAX_WALLS then
		MAX_WALLS = MAX_WALLS * 2
		redim preserve walls(1 to MAX_WALLS) as wall_t ptr
	end if
	NUM_WALLS = NUM_WALLS + 1
	walls(NUM_WALLS) = wall
	function  = NUM_WALLS
end function

'' Set the position of the wall be two points
'' Add it to the list of walls and return an index
function Add_WallByPoints(a as vector, b as vector, byval flags as integer, dv as vector) as integer
	dim t as vector, dr as line_t, tmpwall as wall_t, wall as wall_t ptr
	dr.n = dv
	dr.d = 0

	if (flags and BSP_SOLID) then
		wall = Create_Wall
	else
		wall = @tmpwall
	end if

	wall->aindex = Add_World_Point(a, flags)
	wall->bindex = Add_World_Point(b, flags)
	vSub wall->ab, worldpoints(wall->bindex), worldpoints(wall->aindex)
	VUnitIm wall->ab
	VPerp wall->face.n, wall->ab
	wall->face.d = -VDot(worldpoints(wall->aindex), wall->face.n)

	partition_bsp world_bsp, worldpoints(wall->aindex), worldpoints(wall->bindex), wall->face, flags, dr

	if (flags and BSP_SOLID) then
		function = Add_Wall(wall)
	else
		function = 0
	end if
end function

sub Destroy_All_Walls()
	dim as integer i
	dim as wall_t ptr wall
	for i = 1 to NUM_WALLS
		wall = walls(i)
		if wall then deallocate wall
	next i
	NUM_WALLS = 0
end sub

'' ==========================================================================
'' WORLD
'' ==========================================================================

Sub Create_World()
	world_bsp = create_bsp_tree()
End Sub

Sub Destroy_World()
	Destroy_All_Walls()
	delete_bsp_tree(world_bsp)
	world_bsp = 0
End Sub

function FindWorldCollision(a as vector, d as real, n as vector) as integer
	function = bsp_find_collision(world_bsp, a, d, n)
end function

function GetWorldForceOnPoint(a as vector, f as vector) as integer
	function = bsp_get_force(world_bsp, a, f)
end function

function CheckIsInWorld(a as vector) as integer
	function = bsp_get_side(world_bsp, a)
end function

sub DumpWorldInfo()
	bsp_dump_tree world_bsp
end sub

function CheckVisibility(a as vector, b as vector) as integer
	function = bsp_check_visibility(world_bsp, world_bsp->root, a, b)
end function
